##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

import socket
from .httprequest import HttpRequest
from .httpresponse import HttpResponse
from .server import ClientAddress, Server

class HttpServerMessage:
    def __init__(self, client:socket.socket, address:ClientAddress, data:bytearray) -> None:
        self.client:socket.socket = client
        self.address:ClientAddress = address
        self.data:bytearray = None
        self.req:HttpRequest = None
        self.res:HttpResponse = HttpResponse()
        self.load_request(data)

    def load_request(self, data:bytearray):
        self.data = data
        try:
            self.req = HttpRequest(data.decode('latin-1'), self.client, self.address)
        except:
            self.res.status(403, 'Bad Request')
        return self

class HttpServer(Server):
    def __init__(self, port: int, hostName: str = 'localhost', backlog: int = 8, chunkSize: int = 1024) -> None:
        super().__init__(port, self.on_receive, hostName, backlog, chunkSize)

    def on_handle(self, message:HttpServerMessage) -> bytes:
        return b'HTTP/1.1 200 OK\n\n'

    def on_receive(self, client:socket.socket, address:ClientAddress, data:bytearray) -> bytes:
        if len(data) != 0:
            message = HttpServerMessage(client, address, data)
            return self.on_handle(message)
        return b'HTTP/1.1 200 OK\n\n'


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

from time import time
from typing import Dict, Tuple
import uuid

from ...http.types import Middleware, Next
from ..core.session import Session
from ..core.httprequest import HttpRequest
from ..core.httpresponse import HttpResponse
from ..application import Application

def session(cookieName:str, durationSec:float = 300) -> Middleware:
    sessions:Dict[str, Tuple[Session, float]] = dict()
    lastCheck:float = time()

    def removeExpired():
        nonlocal lastCheck
        if time() - lastCheck > durationSec / 10:
            lastCheck = time()
            expiredSessions = [i for i in sessions if sessions[i][1] < time()]
            for e in expiredSessions:
                del sessions[e]

    def middleware(app:Application, req:HttpRequest, res:HttpResponse, next:Next):
        removeExpired()
        sessionId:str = None
        if cookieName not in req.cookies.values:
            sessionId = str(uuid.uuid4())
            res.set_cookie(cookieName, sessionId, durationSec)
        else:
            sessionId = req.cookies.values[cookieName]

        if sessionId not in sessions:
            sessions[sessionId] = [Session(), durationSec + time()]
        else:
            sessions[sessionId][1] = durationSec + time()

        req.session = sessions[sessionId][0]
        next()
    return middleware


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

import asyncio
import socket
from .common import ClientAddress, OnReceiveCallback

class Server:
    def __init__(self, port:int, onReceive:OnReceiveCallback, hostName:str = 'localhost', backlog:int = 8, chunkSize:int = 1024) -> None:
        self.port = port
        self.hostname = hostName
        self.backlog = backlog
        self.chunkSize = chunkSize
        self.onReceive = onReceive

    def loop(self):
        return asyncio.get_event_loop()

    async def read(self, client:socket.socket, size:int) -> bytes:
        try:
            return await asyncio.wait_for(self.loop().sock_recv(client, size), 0.1)
        except Exception as e:
            return b''

    async def on_accept(self, client:socket.socket, address:ClientAddress):
        data:bytearray = bytearray()
        while True:
            chunk:bytes = await self.read(client, self.chunkSize)
            if len(chunk) == 0:
                break
            data.extend(chunk)
        response:bytes = self.onReceive(client, address, data)
        await self.loop().sock_sendall(client, response)
        client.close()

    async def run(self):
        serverSocket:socket.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        serverSocket.bind((self.hostname, self.port))
        serverSocket.listen(self.backlog)
        serverSocket.setblocking(False)
        
        loop:asyncio.AbstractEventLoop = self.loop()

        while True:
            client:socket.socket = None
            address:ClientAddress = ('', 0)
            client, address = await loop.sock_accept(serverSocket)
            loop.create_task(self.on_accept(client, address))

    def start(self):
        asyncio.run(self.run())


from typing import Dict

def tag(name:str, content:str = None, cls=None, *children, **attributes) -> str:
    result = f'<{name}'
    attributes = attributes or {}
    if cls is not None: attributes['class'] = cls
    attrib:str = ' '.join([f'{i}="{attributes[i]}"' for i in attributes])
    result += attrib
    if len(children) == 0:
        result += '/>'
    else:
        result += '>'
        for child in children:
            result += child
        result += f'</{name}>'

def link(location:str, text:str, cls:str = None):
    return tag('a', text, cls, href=location)


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

import re
from typing import Any, Callable, Dict, List, Tuple

from .core.httprequest import HttpRequest
from .core.httpresponse import HttpResponse
from .types import Callback, CallbackMutator, Middleware, Next

class Router:

    patterns:List[Tuple[str, str]] = [
        (r"{([\w]+)\:int}", r"(?P<\1>[-]?[\\d]+)"), #Int pattern
        (r"{([\w]+)\:float}", r"(?P<\1>[-]?[\\d]+[\\.]?[\\d]?)"), #Float pattern
        (r"{([\w]+)\:re\(([\w\W]+?)\)}", r"(?P<\1>\2)"), #Regex pattern
        (r"{([\w]+)}", r"(?P<\1>[\\w]+)"), #General pattern
    ]

    def __init__(self) -> None:
        self.routes:Dict[str, Dict[str, Callback]] = dict()
        self.calbackMutator:CallbackMutator = lambda a: a

    def normalize_route(self, route:str) -> str:
        if not route.endswith('/'):
            route += '/'
        if not route.startswith('/'):
            route = '/' + route
        while '//' in route:
            route = route.replace('//', '/')  
        while route.endswith('/') and len(route) > 1:
            route = route[:-1]  
        return route
    
    def register_callback(self, httpMethod:str, route:str) -> CallbackMutator:
        ref:Router = self
        route = self.normalize_route(route)
        httpMethod = httpMethod.upper()
        def decorator(callback:Callback) -> Callback:
            callback = self.calbackMutator(callback)
            if httpMethod not in ref.routes:
                ref.routes[httpMethod] = {route:callback}
            else:
                ref.routes[httpMethod][route] = callback
            return callback
        return decorator

    def use_globally(self, mutator:CallbackMutator):
        currentMutator:CallbackMutator = self.calbackMutator
        newMutator:CallbackMutator = lambda c: mutator(currentMutator(c))
        self.calbackMutator = newMutator
        return self

    def get(self, route) -> CallbackMutator:
        return self.register_callback('GET', route)

    def post(self, route) -> CallbackMutator:
        return self.register_callback('POST', route)

    def put(self, route) -> CallbackMutator:
        return self.register_callback('PUT', route)

    def delete(self, route) -> CallbackMutator:
        return self.register_callback('DELETE', route)

    def use_router(self, router, root):
        for method in router.routes:
            for action in router.routes[method]:
                self.register_callback(method,f'{root}{action}')(router.routes[method][action])
        return self

    def serve(self, route:str, content:str, encoding:str):
        fileName:str = sub(sub(route, '/'), '\\')
        inner_callback:Callback = lambda app, req, res : res.file(fileName, content, encoding=encoding)
        self.get(route)(inner_callback)
        return self

    def __getattr__(self, method) -> Callable[[str],CallbackMutator]:
        def inner(route) -> CallbackMutator:
            return self.register_callback(method.upper(), route)
        return inner

    def test_path(self, method:str, path:str) -> Tuple[Callback, Dict[str,str]]:
        callback:Callback = None
        queryParams:Dict[str, str] = dict()
        if method.upper() in self.routes:
            paths:List[str] = self.routes[method.upper()].copy()
            for route in paths:
                pattern:str = route
                for regex in Router.patterns:
                    pattern = re.sub(regex[0], regex[1], pattern)
                match = re.match(f'^{pattern}$', path)
                if match:
                    callback = self.routes[method.upper()][route]
                    queryParams = match.groupdict()
                    break
        return callback, queryParams

    def create_middleware(self, callback:Callback) -> Middleware:
        def inner_callback(app:Any, req:HttpRequest, res:HttpResponse, next:Next):
            res.status(200, 'OK')
            callback(app, req, res)
            next()
        return inner_callback

def sub(data:str, token:str) -> str:
    return data if token not in data else data[data.index(token)+1:]

    


import time
from typing import Dict
from ..core.httprequest import HttpRequest
from ..core.httpresponse import HttpResponse
from ..application import Application
from ...http.types import Next, Middleware, DictProvider

def default_headers(headersProvider: DictProvider) -> Middleware:
    '''
    Sets response default headers. Those headers are determined by the result of the function "headersProvider"
    on every request.
    '''
    def middleware(app: Application, req: HttpRequest, res: HttpResponse, next:Next) -> None:
        headers = headersProvider()
        for h in headers:
            res.headers[h] = headers[h]
        next()
    return middleware

since = time.time()
requestNumber = 0

def build_default_headers(baseHeaders: Dict[str, str] = None) -> DictProvider:
    '''
    Provides a standar function for building default headers.
    '''
    def provider() -> Dict[str, str]:
        global requestNumber
        requestNumber += 1
        headers = baseHeaders or {}
        headers.update({
            "Date": time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime(time.time())),
            "Active-Since": str(since),
            "Epoch": str(time.time()),
            "Powered-By": "Threadsnake beta",
            "Request-Count": requestNumber
        })
        return headers
    return provider



from typing import Dict, List


contentTypes:Dict[str, List[str]] = {
    "text/": ["html", "css"],
    "text/javascript": ["js"],
    "text/html": ["htm"],
    "application/": ["json", "xml", "pdf", "exe"],
    "image/": ["gif", "png", "jpeg", "bmp", "webp"],
    "image/jpeg": ["jpg"],
    "image/x-icon": ["ico"],
    "audio/": ["mpeg", "webm", "ogg", "midi", "wav"],
    "text/plain": ["txt", "*"]
}


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

from base64 import b64decode
from ...http.types import Next
from ..core.httprequest import HttpRequest
from ..core.httpresponse import HttpResponse
from ..application import Application

def authorization(app:Application, req:HttpRequest, res:HttpResponse, next:Next) -> None:
    authKey:str = 'Authorization'
    if authKey in req.headers:
        authType, authValue = req.headers[authKey].split(' ', 1)
        req.authorization[authType] = ''
        if authType == 'Bearer':
            req.authorization[authType] = authValue 
        elif authType == 'Basic':
            user, password = b64decode(authValue.encode()).decode().split(':')
            req.authorization[authType] = {'user':user, 'password':password}
        del req.headers[authKey]
    next()


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

import importlib
import importlib.util
import os
import sys
from types import ModuleType
from typing import List, Tuple
from uuid import uuid4
import __main__

from ..application import Application
from ..router import Router

ConfiguredRoute = Tuple[str, List[Router]]

def import_module(path:str) -> ModuleType:
    if not path.endswith('.py'):
        path += '.py'    
    moduleName:str = os.path.basename(path) + str(uuid4()).replace('-', '')
    spec = importlib.util.spec_from_file_location(moduleName, path)
    module:ModuleType = importlib.util.module_from_spec(spec)
    sys.modules[moduleName] = module
    spec.loader.exec_module(module)
    return module

def routes_to(app:Application, path:str, root:str):
    baseFolder:str = os.path.dirname(__main__.__file__)
    fullSearchPath = os.sep.join([baseFolder, path.replace('/', os.sep)])
    module:ModuleType = import_module(fullSearchPath)
    for property in dir(module):
        router = getattr(module, property)
        if isinstance(router, Router):
            app.use_router(router, root)

def routes_to_folder(app:Application, path:str):
    location:str = os.path.dirname(os.path.abspath(__main__.__file__))
    fullSearchPath = os.sep.join([location, path])
    fullSearchPath = fullSearchPath.replace('/', os.sep).replace('\\', os.sep)
    files = [
        [path, result[0][len(fullSearchPath)+1:].replace('\\', '/'), file[:-3]] 
        for result in os.walk(fullSearchPath) for file in result[2]
        if file.endswith('.py')
    ]
    routes = [
        [
            '/'.join([j for j in i if len(j) > 0]),
            '/'.join([j for j in i[1:] if len(j) > 0])
        ]
        for i in files
    ]
    for route in routes:
        path, root = route
        routes_to(app, path, root)

##def get_routers(moduleName:str) -> List[Router]:
##    routers:List[Router] = []
##    module:ModuleType = importlib.import_module(moduleName)
##    for property in dir(module):
##        if isinstance(getattr(module, property), Router):
##            routers.append(getattr(module, property))
##    return routers
##
##def load_router(path:str) -> ConfiguredRoute:
##    if not path.endswith('.py'):
##        return
##    path = path[:-3]
##    pathParts:List[str] = [i for i in path.replace(os.sep, '.').split('.') if len(i) > 0]
##    module:str = '.'.join(pathParts)
##    path:str = '/'.join(pathParts[1:])
##    return path, get_routers(module)
##
##def get_files(folder:str) -> List[str]:
##    filo = [os.sep.join([j[0], i]) for j in os.walk(folder) for i in j[2]]
##    print(folder, filo)
##    input('')
##
##def load_routes(path:str) -> List[ConfiguredRoute]:
##    baseFolder:str = os.path.dirname(__main__.__file__)
##    fullSearchPath = os.sep.join([baseFolder, path])
##    get_files(fullSearchPath)
##    base:str = os.path.abspath(os.curdir)
##    print(f"{baseFolder} on {base}")
##    configuredRoutes: List[ConfiguredRoute] = []
##    modules:List[str] = [i for i in os.listdir(fullSearchPath) if i.endswith('.py')]
##    input(modules)
##    for module in modules:
##        configuredRoutes.append(load_router(os.sep.join([path, module])))
##    return configuredRoutes
##        
##def configure_routes(app:Application, path:str) -> None:
##    for configuredRoute in load_routes(path):
##        for router in configuredRoute[1]:
##            app.use_router(router, configuredRoute[0])


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

from typing import Dict, List

from .core.httprequest import retrieve_query_pass
from .core.httpserver import HttpServer, HttpServerMessage
from .router import Router
from .types import Callback, Middleware
import __main__
import os

class Application(HttpServer, Router):
    def __init__(self, port: int, hostName: str = 'localhost', backlog: int = 8, chunkSize: int = 1024):
        HttpServer.__init__(self, port, hostName, backlog, chunkSize)
        Router.__init__(self)
        self.stack:List[Middleware] = []
        self.folder:str = os.path.dirname(__main__.__file__)
    
    def configure(self, middleware:Middleware):
        self.stack.append(middleware)
        return self

    def on_handle(self, message: HttpServerMessage) -> bytes:
        if len(message.data) == 0:
            return
        
        stack:List[Middleware] = self.stack.copy()
        
        def next():
            if len(stack) > 0 and not message.res.ended:
                stack.pop()(self, message.req, message.res, next)

        message.req = retrieve_query_pass(message.req)
        callback:Callback = None
        params:Dict[str, str]
        callback, params = self.test_path(message.req.method, message.req.path)
        if callback:
            message.req.params.update(params)
            stack.append(self.create_middleware(callback))

        stack.reverse()
        try:    
            next()
        except Exception as e:
            message.res.end(str(e)).status(500, 'Internal Server Error')

        return message.res.to_bytes()


from typing import List
from ..core.httprequest import HttpRequest
from ..core.httpresponse import HttpResponse
from ..application import Application
from ...http.types import CallbackMutator, Callback, RequestPredicate

def validates_request(predicate:RequestPredicate, onFailMessage:str = None, onFailStatus:int = 400) -> CallbackMutator:
    '''
    Generalizes request validation using a "predicate" function wich receives the current HttpRequestInstance
    and returns a boolean. If the result is false, the client receives an "onFailStatus" status code and an
    "onFailMessage" message. Otherwise the pipeline executes normally.
    '''
    def mutator(middleware:Callback) -> Callback:
        def callback(app:Application, req:HttpRequest, res: HttpResponse) -> None:
            if predicate(req):
                middleware(app, req, res)
            else:
                res.end(onFailMessage or "Bad Request", onFailStatus)
        return callback
    return mutator

def accepts(contentTypes:List[str]) -> CallbackMutator:
    '''
    Validates the HttpRequest content-type against the list of "contentTypes", with are the allowed ones. 
    Returns an UnsuportedMediaType status code if the content-type is not in the list. Otherwise the 
    pipeline executes normally. Delegates to "validates_request".
    '''
    return validates_request(lambda r: r.contentType in contentTypes, onFailStatus=415)

def accepts_json(callback:Callback) -> Callback:
    '''Especialization of accepts wich just allows json requests.'''
    return accepts(['application/json', 'text/json'])(callback)

def requires_parameters(parameters:List[str]) -> CallbackMutator:
    def predicate(req:HttpRequest) -> bool:
        receivedParameters:List[str] = [i for i in req.params]
        return len([i for i in parameters if i not in receivedParameters]) == 0
    message:str = 'Required parameters: ' + ', '.join(parameters)
    return validates_request(predicate, message)


import json
from ..core.httprequest import HttpRequest
from ..core.httpresponse import HttpResponse
from ..application import Application
from ...http.types import Next

def json_body_parser(app:Application, req:HttpRequest, res:HttpResponse, next:Next) -> None:
    '''
    Configures "json" request type. If identifies a valid json in the request body, it sets
    the dict property of the current HttpRequest instance with the json content.
    '''
    if req.contentType in ['application/json', 'text/json']:
        try:
            req.data = json.loads(req.body.strip())
        except:
            res.status(400, 'Bad Request').write("Can't decode json body")
    next()


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

import os
from ...http.types import Middleware, Next
from ..core.httprequest import HttpRequest
from ..core.httpresponse import HttpResponse
from ..application import Application

def static(rootFolder:str = 'static') -> Middleware:
    def middleware(app:Application, req:HttpRequest, res:HttpResponse, next:Next):
        normalizedUrl = req.path.replace('\\', os.sep).replace('..', '')
        filename:str = os.sep.join([app.folder, rootFolder, normalizedUrl])
        if os.path.isfile(filename):
            res.read_file(filename)
        else:
            next()
    return middleware


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

import os
from time import time
from typing import Callable, Dict, List
from uuid import uuid4

from ...http.core.common import map_dictionary
from ...http.types import Middleware, Next
from ..core.httprequest import HttpRequest
from ..core.httpresponse import HttpResponse
from ..application import Application


def save_as_binary(tempFolder:str, data:str):
    tempFileName:str = os.sep.join([tempFolder, str(uuid4()).replace('-', '')+'.tmp'])
    if not os.path.isdir(tempFolder):
        os.mkdir(tempFolder)
    with open(tempFileName,'wb') as f:
        f.write(data.encode('latin-1'))
    return tempFileName

def get_chunked(req:HttpRequest) -> str:
    body:str = req.body
    if 'chunked' in req.headers.get('Transfer-Encoding', '') and '\r\n' in body:
        res = ''
        chunk_i = 1
        while chunk_i > 0:
            split = body.split('\r\n', maxsplit=1)
            chunk, body = split if len(split) == 2 else [split[0], '']
            chunk_i = int(chunk, 16)
            res += body[:chunk_i]
            body = body[chunk_i+2:]
        body = res
    return body

def build_file_decoder(files:Dict[str, float], folder:str, duration:int):
    def decode_body_parameter(parameter:str, req:HttpRequest):
        header, value = parameter.split('\r\n\r\n', maxsplit=1)
        header = header.replace('\r\n', '; ').replace(': ', '=').replace('"', '')
        headerDict = map_dictionary(header, ';', '=')
        if 'filename' in headerDict:
            filename:str = headerDict['filename']
            tempFileLocation:str = save_as_binary(folder, value)
            req.files[filename] = tempFileLocation
            files[tempFileLocation] = time() + duration
        elif 'name' in headerDict:
            req.params[headerDict['name']] = value
    return decode_body_parameter

def remove_old_files(files:Dict[str, float]) -> None:
    removedFiles:List[str] = []
    for file in files:
        if files[file] < time():
            try:
                os.remove(file)
                removedFiles.append(file)
            except:
                print(f'error removing file: {file}')
    for file in removedFiles:
        del files[file]

def multipart_form_data_parser(tempFolder:str = None, filesDurationSec:int = 30) -> Middleware:
    files:Dict[str, float] = {}
    tempFolder = tempFolder or os.sep.join(['.', 'temp'])
    decoder:Callable[[str, HttpRequest], None] = None
    decoder = build_file_decoder(files, tempFolder, filesDurationSec)
    def middleware(app:Application, req:HttpRequest, res:HttpResponse, next:Next) -> None:
        if 'multipart/form-data' in req.contentType:
            boundary:str = ''
            req.contentType, boundary = [i.strip() for i in req.contentType.split(';', 1)]
            boundary = boundary.strip().replace('boundary=', '--').replace('"', '')
            body:str = get_chunked(req)
            bodyParameters:List[str] = [i.strip() for i in body.split(boundary)]
            for param in bodyParameters:
                if param == '--' or len(param) == 0:
                    continue
                decoder(param, req)
        remove_old_files(files)
        next()
    return middleware



##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

import json
from socket import socket
from typing import Any, Dict, List

from .session import Session

from .common import decode_querystring, map_dictionary
from .constants import HEADER_CONTENT_TYPE, HEADER_COOKIES
from .server import ClientAddress

class HttpQuery:
    def __init__(self) -> None:
        self.url = ''
        self.querystring = ''
        self.path = ''
    
    def load_query(self, url:str) -> Dict[str, str]:
        self.url = url
        self.path = url
        if '?' in url:
            self.path, url = url.split('?', 1)
            self.querystring = decode_querystring(url)
            return map_dictionary(self.querystring, '&', '=')
        return dict()

class Cookies:
    def __init__(self, values:Dict[str, str] = None) -> None:
        self.values = values or dict()

    def load_cookies(self, raw:str):
        self.values = map_dictionary(raw, ';', '=')
        return self

class HttpRequest(HttpQuery):
    def __init__(self, raw:str, client:socket, clientAddress:ClientAddress) -> None:
        HttpQuery.__init__(self)
        self.client = client
        self.clientAddress = clientAddress
        self.headers:Dict[str, str] = {}
        self.params:Dict[str, str] = {}
        self.cookies:Cookies = Cookies()
        self.contentType:str = ''
        self.files:Dict[str, str] = {}
        self.authorization = {}
        self.session:Session = None
        self.data:Any = None
        self.raw = ''
        self.body:str = ''
        self.method:str = ''
        self.httpVersion:str = ''
        self.load(raw)

    def load(self, raw:str):
        self.raw = raw
        requestParts:List[str] = raw.split('\r\n\r\n', maxsplit=1)
        statusAndHeaders:List[str] = requestParts[0].split('\r\n')
        self.load_status(statusAndHeaders[0])
        self.load_headers(statusAndHeaders[1:])
        if len(requestParts) > 1:
            self.load_body(requestParts[1])

    def load_status(self, statusLine:str):
        status:List[str] = statusLine.split(' ')
        self.method = status[0]
        self.httpVersion = status[2] if len(status) > 2 else 'HTTP/1.1'
        if len(status) > 1:
            self.params = self.load_query(status[1])


    def load_headers(self, headers:List[str]):
        headerPairs:List[List[str]] = [
            [j.strip() for j in i.split(':', 1)] 
            for i in headers
            if len(i.split(':')) == 2
        ]
        self.headers = {header[0]:header[1] for header in headerPairs}
        if HEADER_COOKIES in self.headers:
            self.cookies = Cookies().load_cookies(self.headers[HEADER_COOKIES])
        if HEADER_CONTENT_TYPE in self.headers:
            self.contentType = self.headers[HEADER_CONTENT_TYPE]

    def load_body(self, body:str):
        self.body = body

    def json(self):
        try:
            return json.loads(self.body)
        except:
            return None

def retrieve_query_pass(req:HttpRequest) -> HttpRequest:
    if ':' in req.path:
        path:List[str] = req.path.split('/')
        req.path = '/'.join([i for i in path if len(i.split(':', 1)) != 2])
        req.params.update({i[0]:i[1] for i in path if len(i.split(':', 1)) == 2})
    return req


import os
import setuptools

def readfile(filename):
    with open(filename, 'r', encoding='latin1') as f:
        return f.read()

setuptools.setup(    
    name="threadsnake",
    version=readfile('version.txt'),
    author="Erick Fernando Mora Ramirez",
    author_email="erickfernandomoraramirez@gmail.com",
    description="A tiny experimental server-side express-like library",
    long_description=readfile('README.MD'),
    long_description_content_type="text/markdown",
    url="https://github.com/LostSavannah/threadsnake",
    project_urls={
        "Bug Tracker": "https://dev.moradev.dev/threadsnake/issues/",
        "Documentation": "https://dev.moradev.dev/threadsnake/documentation/",
        "Examples": "https://dev.moradev.dev/threadsnake/examples/",
    },
    package_data={
        "":["*.txt"]
    },
    classifiers=[
    "Programming Language :: Python :: 3",
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6"
)


from typing import Dict

responseCodes:Dict[int, str] = {
    100 : "Continue",
	101 : "Switching Protocols",
	102 : "Processing (WebDAV)",
	103 : "Early Hints",
	200 : "OK",
	201 : "Created",
	202 : "Accepted",
	203 : "Non-Authoritative Information",
	200 : "OK response is preferred to this status",
	204 : "No Content",
	205 : "Reset Content",
	206 : "Partial Content",
	207 : "Multi-Status (WebDAV)",
	208 : "Already Reported (WebDAV)",
	226 : "IM Used (HTTP Delta encoding)",
	300 : "Multiple Choices",
	301 : "Moved Permanently",
	302 : "Found",
	303 : "See Other",
	304 : "Not Modified",
	305 : "Use Proxy Deprecated",
	306 : "unused",
	307 : "Temporary Redirect",
	302 : "Found HTTP response code",
	308 : "Permanent Redirect",
	301 : "Moved Permanently HTTP response code",
	400 : "Bad Request",
	401 : "Unauthorized",
	402 : "Payment Required Experimental",
	403 : "Forbidden",
	401 : "Unauthorized",
	404 : "Not Found",
	403 : "Forbidden to hide the existence of a resource from an unauthorized client",
	405 : "Method Not Allowed",
	406 : "Not Acceptable",
	407 : "Proxy Authentication Required",
	401 : "Unauthorized but authentication is needed to be done by a proxy",
	408 : "Request Timeout",
	409 : "Conflict",
	410 : "Gone",
	411 : "Length Required",
	412 : "Precondition Failed",
	413 : "Payload Too Large",
	414 : "URI Too Long",
	415 : "Unsupported Media Type",
	416 : "Range Not Satisfiable",
	417 : "Expectation Failed",
	418 : "I'm a teapot",
	421 : "Misdirected Request",
	422 : "Unprocessable Entity (WebDAV)",
	423 : "Locked (WebDAV)",
	424 : "Failed Dependency (WebDAV)",
	425 : "Too Early Experimental",
	426 : "Upgrade Required",
	426 : "response to indicate the required protocol(s)",
	428 : "Precondition Required",
	429 : "Too Many Requests",
	431 : "Request Header Fields Too Large",
	451 : "Unavailable For Legal Reasons",
	500 : "Internal Server Error",
	501 : "Not Implemented",
	502 : "Bad Gateway",
	503 : "Service Unavailable",
	504 : "Gateway Timeout",
	505 : "HTTP Version Not Supported",
	506 : "Variant Also Negotiates",
	507 : "Insufficient Storage (WebDAV)",
	508 : "Loop Detected (WebDAV)",
	510 : "Not Extended",
	511 : "Network Authentication Required"
}


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

from typing import Any, Dict

class Session:
    def __init__(self) -> None:
        self.values:Dict[str,Any] = dict()
    
    def set(self, key:str, value:Any):
        self.values[key] = value

    def has(self, key:str) -> bool:
        return key in self.values

    def get(self, key:str, default:Any) -> Any:
        if not self.has(key):
            self.set(key, default)
        return self.values[key]


from ..core.httprequest import HttpRequest
from ..core.httpresponse import HttpResponse
from ..application import Application
from ...http.types import Next

def cors(app: Application, req: HttpRequest, res: HttpResponse, next:Next) -> None:
    '''Configures CORS.'''
    res.set_header('Access-Control-Allow-Origin', "*")
    next()


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

from ...http.types import Next
from ..core.httprequest import HttpRequest
from ..core.httpresponse import HttpResponse
from ..application import Application
from ..core.common import decode_querystring, map_dictionary

def body_parser(app:Application, req:HttpRequest, res:HttpResponse, next:Next) -> None:
    if req.contentType != None and 'x-www-form-urlencoded' in req.contentType:
        rawParams:str = decode_querystring(req.body.strip())
        req.params.update(map_dictionary(rawParams, '&', '='))
    next()


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

import re
import socket
import time
from typing import Any, Callable, Dict, List, Tuple

from .values.contenttypes import contentTypes
from .values.responsecodes import responseCodes

ClientAddress = Tuple[str, int]
OnReceiveCallback = Callable[[Any, socket.socket, ClientAddress], bytes]


def get_content_type(path:str):
    contentType = 'text/plain'
    if '.' in path:
        extension:str = path.split(".")[-1:][0]
        for i in contentTypes:
            if extension in contentTypes[i]:
                contentType = i
                if contentType.endswith("/"):
                    contentType += extension
                break
    return contentType

def map_dictionary(data:str, rowSeparator:str, keySeparator:str) -> Dict[str, str]:
    table:List[List[str]] = [
        [j.strip() for j in i.split(keySeparator, 1)]
        for i in data.split(rowSeparator) 
        if len(i.split(keySeparator)) == 2
    ]
    return {row[0]:row[1] for row in table}

def decode_querystring(data:str) -> Dict[str, str]:
    data = data.replace('+', ' ').replace('%20', ' ')#space
    data = data.replace('%2B', '+')#space
    data = data.replace('%7E', '~')#space
    references = [(i, chr(int(re.findall(r'[\d]+', i)[0]))) for i in re.findall(r'&#[\d]+;', data)]
    for ref in references:
        data = data.replace(ref[0], ref[1])
    return data

def get_cookie_expiration_UTC(durationSec:float) -> str:
    return time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime(time.time() + durationSec))

def get_port(port:int, max_port=65535) -> int:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        while port <= max_port:
            try:
                sock.bind(('', port))
                sock.close()
                return port
            except OSError:
                port += 1
        else:
            raise IOError('no free ports')

def get_status_text_from_status_code(statusCode:int) -> str:
    return responseCodes.get(statusCode, 'Unknown')


from .authorization import authorization
from .bodyparser import body_parser
from .cors import cors
from .defaultheaders import default_headers, build_default_headers
from .jsonbodyparser import json_body_parser
from .multipartformdataparser import multipart_form_data_parser
from .session import session
from .static import static
from .timemeassure import time_measure


from time import time

from ..core.httprequest import HttpRequest
from ..core.httpresponse import HttpResponse
from ..application import Application
from ...http.types import Next

def time_measure(app:Application, req:HttpRequest, res:HttpResponse, next:Next) -> None:
    '''
    Ideally measures the time than all pipeline takes to execute.
    '''
    startTime = time()
    next()
    interval = (time() - startTime) * 1000
    res.headers['RTT'] = str(interval)


##    Threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

import json
from typing import Any, Dict, List
from .common import get_content_type, get_cookie_expiration_UTC, get_status_text_from_status_code
from .constants import HEADER_CONTENT_DISPOSITION, HEADER_CONTENT_LENGTH, HEADER_CONTENT_TYPE, HEADER_LOCATION, LINESEP

class ResponseHead:
    def __init__(self) -> None:
        self.httpVersion:str = 'HTTP/1.1'
        self.responseStatus:int = 404
        self.responseStatusText:str = 'NotFound'
        self.headers:Dict[str, List[str]] = dict()
    
    def append_header(self, name:str, value:str):
        if name not in self.headers:
            self.headers[name] = []
        self.headers[name].append(value)
    
    def set_header(self, name:str, value:str):
        self.headers[name] = [value]

    def get_header(self, name:str) -> str:
        if name in self.headers:
            return self.headers[name][0]
        return None
 
    def get_headers(self, name:str) -> List[str]:
        if name in self.headers:
            return self.headers[name]
        return None

    def remove_header(self, name:str) -> str:
        if name in self.headers:
            del self.headers[name]

    def set_cookie(self, name:str, value:str, durationSec:float = None, domain:str = None, path:str = None):
        cookieString = [value]
        if durationSec != None: cookieString.append(f'Expires={get_cookie_expiration_UTC(durationSec)}')
        if domain is not None: cookieString.append(f'Domain={domain}')
        if path is not None: cookieString.append(f'Path={path}')
        cookieHeader = f'{name}=' + '; '.join(cookieString)
        self.append_header('Set-Cookie', cookieHeader)

    def to_string(self) -> str:
        result:str = f'{self.httpVersion} {self.responseStatus} {self.responseStatusText} {LINESEP}'
        for header in self.headers:
            for headerValue in self.headers[header]:
                result += f'{header}: {headerValue}{LINESEP}'
        return result

class HttpResponse(ResponseHead):
    def __init__(self) -> None:
        ResponseHead.__init__(self)
        self.body:str = ''
        self.encoding:str = 'latin-1'
        self.ended:bool = False
    
    def end(self, data:str = None, status:int = None):
        self.ended = True
        return self.status(status).write(data)

    def status(self, responseCode:int = None, responseText:str = None):
        self.responseStatus = responseCode or self.responseStatus
        responseText = responseText or get_status_text_from_status_code(self.responseStatus)
        self.responseStatusText = responseText
        return self

    def append_header(self, name: str, value: str):
        ResponseHead.append_header(self, name, value)
        return self

    def set_header(self, name: str, value: str):
        ResponseHead.set_header(self, name, value)
        return self

    def set_cookie(self, name: str, value: str, durationSec: float = None, domain: str = None, path: str = None):
        ResponseHead.set_cookie(self, name, value, durationSec, domain, path)
        return self

    def content_disposition(self, value:str, fileName:str = None):
        if fileName is not None:
            value += f'; filename="{fileName}"'
        return self.set_header(HEADER_CONTENT_DISPOSITION, value)

    def content_type(self, value:str):
        return self.set_header(HEADER_CONTENT_TYPE, value)

    def redirect(self, value:str):
        return self.set_header(HEADER_LOCATION, value)

    def set_encoding(self, value:str):
        self.encoding = value
        return self

    def write(self, data:str):
        self.body += data
        return self

    def json(self, data:Any):
        return self\
        .write(json.dumps(data))\
        .content_type('application/json')\
        .status(200, 'OK')

    def html(self, data:str):
        return self\
        .write(data)\
        .content_type('text/html')\
        .status(200, 'OK')

    def read_file(self, fileName:str, encoding:str = None):
        self.set_encoding(encoding or self.encoding)
        with open(fileName, 'r', encoding=self.encoding) as f:
            self.write(f.read())
        return self.status(200, 'OK').content_type(get_content_type(fileName))

    def file(self, fileName:str, data:str, contentType:str = None, encoding:str = None):
        self.content_disposition('attachment', fileName)
        self.set_encoding(encoding or self.encoding)
        self.status(200, 'OK')
        if contentType is not None:
            self.content_type(contentType)
        self.body = data

    def download(self, path:str, fileName:str, contentType:str = None, encoding:str = None):
        self.fileName('attachment', fileName)
        self.body = ""
        return self.read_file(path, encoding)

    def to_bytes(self)->bytes:
        return str(self).encode(self.encoding)

    def __str__(self) -> str:
        if HEADER_LOCATION not in self.headers:
            self.set_header(HEADER_CONTENT_LENGTH, len(self.body))
        return ResponseHead.to_string(self) + LINESEP + self.body
